import * as React from 'react';
import WaveSurfer from 'wavesurfer.js';
import {AudioFile, getMimeTypeDescription} from '../../utils/blob';
import {formatFileSize, formatNumeral, formatSecondsDurationPrecise} from '../../utils/numeral';

export interface VideoFilePreviewProps {
  file: AudioFile;
}

export const AudioFilePreview: React.FC<VideoFilePreviewProps> = ({
  file: f,
}) => {
  const mediaContainerRef = React.useRef<HTMLDivElement>(null);
  const mediaControllerRef = React.useRef<WaveSurfer>(null);
  const [isPlaying, setIsPlaying] = React.useState(false);

  React.useEffect(() => {
    if (!mediaControllerRef.current) {
      return;
    }

    if (isPlaying) {
      void mediaControllerRef.current.play();
      return;
    }

    mediaControllerRef.current.pause();
  }, [isPlaying]);

  React.useEffect(() => {
    if (!mediaControllerRef.current) {
      return;
    }

    if (!mediaContainerRef.current) {
      return;
    }

    if (typeof f.metadata?.previewUrl !== 'string') {
      return;
    }

    mediaContainerRef.current.innerHTML = '';

    if (f.type === 'audio/mid') {
      return;
    }

    const mediaControllerRefMutable = mediaControllerRef as React.MutableRefObject<WaveSurfer>;
    mediaControllerRefMutable.current = WaveSurfer.create({
      container: mediaContainerRef.current,
      url: f.metadata.previewUrl,
      cursorWidth: 0,
      height: mediaContainerRef.current.offsetHeight,
      barWidth: 2,
      barGap: 2,
      barRadius: 1,
    });

    mediaControllerRefMutable.current.on('finish', () => {
      setIsPlaying(false);
      mediaControllerRefMutable.current.seekTo(0);
    });

    return () => {
      if (!mediaControllerRefMutable.current) {
        return;
      }

      mediaControllerRefMutable.current.destroy();
    }
  }, [f, mediaContainerRef, mediaControllerRef]);

  const playMedia = () => {
    setIsPlaying((p) => !p);
  };

  return (
    <div className="flex flex-col gap-4 w-full h-full relative">
      <div className="h-2/5 flex-shrink-0 cursor-pointer relative">
        <div
          ref={mediaContainerRef}
          className="relative h-full w-full"
        />
        <div className="absolute bottom-0 left-0 z-[2]">
          <button
            onClick={playMedia}
          >
            {isPlaying ? '⏸' : '▶'}
          </button>
        </div>
      </div>
      <dl className="h-3/5 flex-shrink-0 m-0 flex flex-col items-end" data-testid="infoBox">
        <div className="w-full">
          <dt className="sr-only">
            Name
          </dt>
          <dd
            className="m-0 w-full text-ellipsis overflow-hidden"
            title={f.name}
          >
            {f.name}
          </dd>
        </div>
        <div className="w-full">
          <dt className="sr-only">
            Type
          </dt>
          <dd
            title={f.type}
            className="m-0 w-full text-ellipsis overflow-hidden"
          >
            {getMimeTypeDescription(f.type)}
          </dd>
        </div>
        <div className="w-full">
          <dt className="sr-only">
            Size
          </dt>
          <dd
            className="m-0 w-full text-ellipsis overflow-hidden"
            title={`${formatNumeral(f.size ?? 0)} bytes`}
          >
            {formatFileSize(f.size)}
          </dd>
        </div>
        {
          typeof f.metadata?.duration === 'number'
          && (
            <div className="w-full">
              <dt className="sr-only">
                Duration
              </dt>
              <dd
                className="m-0 w-full text-ellipsis overflow-hidden"
                title={`${formatNumeral(f.metadata.duration ?? 0)} seconds`}
              >
                {formatSecondsDurationPrecise(f.metadata.duration)}
              </dd>
            </div>
          )
        }
      </dl>
    </div>
  );
};
