import * as React from 'react';
import {getMimeTypeDescription, VideoFile} from '../../utils/blob';
import {formatFileSize, formatNumeral, formatSecondsDurationConcise} from '../../utils/numeral';

export interface VideoFilePreviewProps {
  file: VideoFile;
}

interface UseVideoControlsOptions {
  mediaControllerRef: React.Ref<HTMLVideoElement>;
}

const useVideoControls = ({
  mediaControllerRef,
}: UseVideoControlsOptions) => {
  const seekRef = React.useRef<HTMLInputElement>(null);
  const volumeRef = React.useRef<HTMLInputElement>(null);
  const [isPlaying, setIsPlaying] = React.useState(false);
  const [isSeeking, setIsSeeking] = React.useState(false);
  const [currentTimeDisplay, setCurrentTimeDisplay] = React.useState<number>();
  const [seekTimeDisplay, setSeekTimeDisplay] = React.useState<number>();
  const [durationDisplay, setDurationDisplay] = React.useState<number>();
  const [isSeekTimeCountingDown, setIsSeekTimeCountingDown] = React.useState(false);

  const refreshControls: React.ReactEventHandler<HTMLVideoElement> = (e) => {
    const { currentTarget: mediaController } = e;
    const { current: seek } = seekRef;
    if (!seek) {
      return;
    }

    setCurrentTimeDisplay(mediaController.currentTime);
    setDurationDisplay(mediaController.duration);
  };

  const playMedia = React.useCallback(() => {
    setIsPlaying((p) => !p);
  }, []);

  const startSeek: React.MouseEventHandler<HTMLInputElement> = React.useCallback(() => {
    setIsSeeking(true);
  }, []);

  const doSetSeek = (thisElement: HTMLInputElement, mediaController: HTMLVideoElement) => {
    mediaController.currentTime = thisElement.valueAsNumber;
  };

  const setSeek: React.ChangeEventHandler<HTMLInputElement> = React.useCallback((e) => {
    if (!(typeof mediaControllerRef === 'object' && mediaControllerRef !== null)) {
      return;
    }

    const { current: mediaController } = mediaControllerRef;
    if (!mediaController) {
      return;
    }

    const { currentTarget: thisElement } = e;
    setSeekTimeDisplay(thisElement.valueAsNumber);

    if (isSeeking) {
      return;
    }

    doSetSeek(thisElement, mediaController);
  }, [mediaControllerRef, isSeeking]);

  const endSeek: React.MouseEventHandler<HTMLInputElement> = React.useCallback((e) => {
    if (!(typeof mediaControllerRef === 'object' && mediaControllerRef !== null)) {
      return;
    }

    const { current: mediaController } = mediaControllerRef;

    if (!mediaController) {
      return;
    }

    const { currentTarget: thisElement } = e;
    setIsSeeking(false);
    doSetSeek(thisElement, mediaController);
  }, [mediaControllerRef]);

  const resetVideo: React.ReactEventHandler<HTMLVideoElement> = React.useCallback((e) => {
    const videoElement = e.currentTarget;
    setIsPlaying(false);
    videoElement.currentTime = 0;
  }, []);

  const updateSeekFromPlayback: React.ReactEventHandler<HTMLVideoElement> = React.useCallback((e) => {
    if (!seekRef.current) {
      return;
    }

    if (isSeeking) {
      return;
    }

    const { current: seek } = seekRef;

    if (!seek) {
      return;
    }

    const videoElement = e.currentTarget;
    const currentTime = videoElement.currentTime;
    setCurrentTimeDisplay(currentTime);
    seek.value = String(currentTime);
  }, [isSeeking]);

  const adjustVolume: React.ChangeEventHandler<HTMLInputElement> = React.useCallback((e) => {
    if (!(typeof mediaControllerRef === 'object' && mediaControllerRef !== null)) {
      return;
    }

    if (!mediaControllerRef.current) {
      return;
    }
    const { value } = e.currentTarget;
    mediaControllerRef.current.volume = Number(value);
  }, [mediaControllerRef]);

  const toggleSeekTimeCountMode: React.MouseEventHandler<HTMLButtonElement> = React.useCallback((e) => {
    e.preventDefault();
    setIsSeekTimeCountingDown((b) => !b);
  }, []);

  React.useEffect(() => {
    if (!(typeof mediaControllerRef === 'object' && mediaControllerRef !== null)) {
      return;
    }

    if (!mediaControllerRef.current) {
      return;
    }

    if (isPlaying) {
      void mediaControllerRef.current.play();
      return
    }

    mediaControllerRef.current.pause();
  }, [isPlaying, mediaControllerRef]);

  React.useEffect(() => {
    if (!seekRef.current) {
      return;
    }

    const { current: seek } = seekRef;
    if (!seek) {
      return;
    }

    seek.value = String(currentTimeDisplay);
  }, [currentTimeDisplay]);

  React.useEffect(() => {
    if (!seekRef.current) {
      return;
    }

    const { current: seek } = seekRef;
    if (!seek) {
      return;
    }

    seek.max = String(durationDisplay);
  }, [durationDisplay]);

  React.useEffect(() => {
    if (!volumeRef.current) {
      return;
    }

    if (!(typeof mediaControllerRef === 'object' && mediaControllerRef !== null)) {
      return;
    }

    if (!mediaControllerRef.current) {
      return;
    }

    const { current: mediaController } = mediaControllerRef;
    const { current: volume } = volumeRef;
    volume.value = String(mediaController.volume);
  }, [mediaControllerRef]);

  return React.useMemo(() => ({
    seekRef,
    volumeRef,
    isPlaying,
    refreshControls,
    adjustVolume,
    resetVideo,
    startSeek,
    endSeek,
    setSeek,
    updateSeekFromPlayback,
    playMedia,
    durationDisplay,
    currentTimeDisplay,
    seekTimeDisplay,
    isSeeking,
    isSeekTimeCountingDown,
    toggleSeekTimeCountMode,
  }), [
    isPlaying,
    isSeeking,
    adjustVolume,
    playMedia,
    resetVideo,
    startSeek,
    endSeek,
    setSeek,
    updateSeekFromPlayback,
    durationDisplay,
    currentTimeDisplay,
    seekTimeDisplay,
    isSeekTimeCountingDown,
    toggleSeekTimeCountMode,
  ]);
};

export const VideoFilePreview = React.forwardRef<HTMLVideoElement, VideoFilePreviewProps>(({
  file: f,
}, forwardedRef) => {
  const defaultRef = React.useRef<HTMLVideoElement>(null);
  const mediaControllerRef = forwardedRef ?? defaultRef;
  const {
    seekRef,
    volumeRef,
    isPlaying,
    adjustVolume,
    playMedia,
    resetVideo,
    startSeek,
    endSeek,
    setSeek,
    updateSeekFromPlayback,
    refreshControls,
    durationDisplay = 0,
    currentTimeDisplay = 0,
    seekTimeDisplay = 0,
    isSeeking,
    isSeekTimeCountingDown,
    toggleSeekTimeCountMode,
  } = useVideoControls({
    mediaControllerRef,
  });

  const finalSeekTimeDisplay = isSeekTimeCountingDown ? (durationDisplay - seekTimeDisplay) : seekTimeDisplay;
  const finalCurrentTimeDisplay = isSeekTimeCountingDown ? (durationDisplay - currentTimeDisplay) : currentTimeDisplay;

  return (
    <div className="flex gap-4 w-full h-full relative">
      <div className={`h-full w-1/3 flex-shrink-0`}>
        {
          typeof f.metadata?.previewUrl === 'string'
          && (
            <div
              className="w-full h-full bg-black flex flex-col items-stretch"
              data-testid="preview"
            >
              <div className="w-full flex-auto relative">
              <video
                className="absolute w-full h-full top-0 left-0 block object-center object-contain flex-auto"
                ref={mediaControllerRef as React.RefObject<HTMLVideoElement>}
                onLoadedMetadata={refreshControls}
                onDurationChange={refreshControls}
                onEnded={resetVideo}
                onTimeUpdate={updateSeekFromPlayback}
              >
                <source
                  src={f.metadata.previewUrl}
                  type={f.type}
                />
              </video>
              </div>
              <div className="w-full flex-shrink-0 h-10 flex gap-4">
                <button
                  onClick={playMedia}
                  className="w-10 h-full flex-shrink-0"
                  type="button"
                >
                  {isPlaying ? '⏸' : '▶'}
                </button>
                <div className="flex-auto w-full flex items-center gap-1 font-mono text-sm relative">
                  <button
                    className="absolute overflow-hidden w-12 opacity-0 h-10"
                    onClick={toggleSeekTimeCountMode}
                    title="Toggle Seek Time Count Mode"
                    type="button"
                  >
                    Toggle Seek Time Count Mode
                  </button>
                  <span
                    className="font-mono before:block before:content-[attr(title)] contents"
                    title={
                      `${
                        isSeekTimeCountingDown ? '-' : '+'
                      }${
                        isSeeking
                          ? formatSecondsDurationConcise(finalSeekTimeDisplay)
                          : formatSecondsDurationConcise(finalCurrentTimeDisplay)
                      }`
                    }
                  >
                    <input
                      type="range"
                      className="flex-auto w-full"
                      ref={seekRef}
                      onMouseDown={startSeek}
                      onMouseUp={endSeek}
                      onChange={setSeek}
                      defaultValue="0"

                    />
                  </span>
                  <span>
                    {formatSecondsDurationConcise(durationDisplay)}
                  </span>
                </div>
                <input
                  type="range"
                  ref={volumeRef}
                  max={1}
                  min={0}
                  onChange={adjustVolume}
                  step="any"
                  className="flex-shrink-0 w-12"
                  defaultValue="1"
                  title="Volume"
                />
              </div>
            </div>
          )
        }
      </div>
      <dl className="w-2/3 flex-shrink-0 m-0" data-testid="infoBox">
        <div className="w-full">
          <dt className="sr-only">
            Name
          </dt>
          <dd
            className="m-0 w-full text-ellipsis overflow-hidden"
            title={f.name}
          >
            {f.name}
          </dd>
        </div>
        <div className="w-full">
          <dt className="sr-only">
            Type
          </dt>
          <dd
            title={f.type}
            className="m-0 w-full text-ellipsis overflow-hidden"
          >
            {getMimeTypeDescription(f.type, f.name)}
          </dd>
        </div>
        <div className="w-full">
          <dt className="sr-only">
            Size
          </dt>
          <dd
            className="m-0 w-full text-ellipsis overflow-hidden"
            title={`${formatNumeral(f.size ?? 0)} bytes`}
          >
            {formatFileSize(f.size)}
          </dd>
        </div>
        {
          typeof f.metadata?.width === 'number'
          && typeof f.metadata?.height === 'number'
          && (
            <div>
              <dt className="sr-only">
                Pixel Dimensions
              </dt>
              <dd
                className="m-0 w-full text-ellipsis overflow-hidden"
              >
                {formatNumeral(f.metadata.width)}&times;{formatNumeral(f.metadata.height)} pixels
              </dd>
            </div>
          )
        }
      </dl>
    </div>
  );
});

VideoFilePreview.displayName = 'VideoFilePreview';
